{-# LANGUAGE BangPatterns #-}

import System.Environment (getArgs)
import System.IO (readFile)
import Data.Array (array, listArray, assocs)
import Data.Foldable (maximumBy)
import Data.Ord (comparing)
import Control.Parallel.Strategies (parMap,parListChunk, rdeepseq)
import Othello.GameLogic (
    GameSetup(..), Player(..), initialGameState, Board(..), DiscState(..),
    movesForPlayer, applyMove, Coordinate, opposingPlayer 
    ) 
import GHC.Conc (numCapabilities)
import Data.List.Split (chunksOf)

-- Parse a custom board state from a string
parseCustomBoard :: String -> Int -> Board
parseCustomBoard input dim = Board {
    grid = listArray ((0, 0), (dim-1, dim-1)) discStates,
    boardDim = dim
  }
  where
    rows = lines input
    discStates = concatMap parseRow rows
    parseRow row = map parseDiscState (words row)
    parseDiscState "E" = Empty
    parseDiscState "R" = Placed Red
    parseDiscState "B" = Placed Blue
    parseDiscState _   = error "Invalid disc state in custom board. Use 'E', 'R', or 'B'."


-- Define a simple entry point for the program
main :: IO ()
main = do
    -- Get command-line arguments
    args <- getArgs

    -- Parse depth arguments
    let (depth, parallelDepth, inputFile) = case args of
            (d:p:file:_) -> (read d, read p, Just file)
            (d:p:_)      -> (read d, read p, Nothing)
            _            -> error "Usage: <depth> <parallelDepth> [gameboard file]"

    -- Load the custom board or use the default initial game state
    gameSetup <- case inputFile of
        Just file -> do
            content <- readFile file
            let customBoard = parseCustomBoard content 8 -- Assume an 8x8 board for simplicity
            return GameSetup {
                board = customBoard,
                aiPlays = [Red, Blue],
                searchDepth = depth
            }
        Nothing -> return (initialGameState 8 [Red, Blue] depth)

    putStrLn $ "Using minimax depth: " ++ show depth
    putStrLn $ "Parallelizing at depth: " ++ show parallelDepth

    let currentPlayer = Red
    
    let moves = getAIMove gameSetup currentPlayer depth parallelDepth
    if null moves
        then putStrLn "No valid moves available."
        else do
            let selectedMove = head moves
            let updatedBoard = applyMove (board gameSetup) currentPlayer [selectedMove]
            putStrLn $ "Next move: " ++ show selectedMove
            putStrLn "Updated board state:"
            print updatedBoard
            
-- Define a custom AI move function using alpha-beta pruning
getAIMove :: GameSetup -> Player -> Int -> Int -> [Coordinate]
getAIMove setup player depth parallelDepth
    | null possibleMoves = [] -- No moves available
    | otherwise = fst $ maximumBy (comparing snd) evaluatedMoves
  where
    possibleMoves = movesForPlayer (board setup) player

    evaluatedMoves =
        if depth == parallelDepth
        then parallelEvaluate possibleMoves
        else sequentialEvaluate possibleMoves

    sequentialEvaluate moves =
        [(move, minimax (applyMove (board setup) player move) (opposingPlayer player) (depth - 1) parallelDepth minBound maxBound)
        | move <- moves]

    parallelEvaluate moves =
        let chunkSize = max 1 (length moves `div` numCapabilities) -- Split moves into chunks
            eval move = (move, minimax (applyMove (board setup) player move) (opposingPlayer player) (depth - 1) parallelDepth minBound maxBound)
        in concat $ parMap rdeepseq (map eval) (chunksOf chunkSize moves)
        
-- Minimax algorithm with alpha-beta pruning
minimax :: Board -> Player -> Int -> Int -> Int -> Int -> Int
minimax board player depth parallelDepth alpha beta
    | depth == 0 || null possibleMoves = evaluateBoard board player
    | player == maximizingPlayer = maximize alpha beta possibleMoves
    | otherwise = minimize alpha beta possibleMoves
  where
    possibleMoves = movesForPlayer board player
    maximizingPlayer = Red

    maximize :: Int -> Int -> [[Coordinate]] -> Int
    maximize a b [] = a
    maximize a b (move:moves)
        | a' >= b  = a'
        | otherwise = maximize a' b moves
      where
        a' = max a (nextEval move a b)

    minimize :: Int -> Int -> [[Coordinate]] -> Int
    minimize a b [] = b
    minimize a b (move:moves)
        | a >= b' = b'
        | otherwise = minimize a b' moves
      where
        b' = min b (nextEval move a b)

    nextEval move a b
        | depth == parallelDepth =
            let results = parMap rdeepseq eval possibleMoves
            in if player == maximizingPlayer then maximum results else minimum results
        | otherwise =
            minimax (applyMove board player move) (opposingPlayer player) (depth - 1) parallelDepth a b
        where
            eval move = minimax (applyMove board player move) (opposingPlayer player) (depth - 1) parallelDepth a b



-- Board evaluation
evaluateBoard :: Board -> Player -> Int
evaluateBoard board player = scoreForPlayer - scoreForOpponent
  where
    scoreForPlayer = length [pos | (pos, state) <- assocs (grid board), state == Placed player]
    scoreForOpponent = length [pos | (pos, state) <- assocs (grid board), state == Placed (opposingPlayer player)]


