module Main where


import Data.List.Split (chunksOf)
import GHC.Conc (numCapabilities)
import System.Environment (getArgs)
import System.IO (readFile)
import Data.Array (array, listArray, assocs)
import Control.Parallel.Strategies (parMap, rdeepseq, parBuffer, using)
import Data.List (maximumBy)
import Data.Ord (comparing)
import Othello.GameLogic (
    GameSetup(..), Player(..), initialGameState, Board(..), DiscState(..),
    movesForPlayer, applyMove, Coordinate, opposingPlayer
    )

-- Parse a custom board state from a string
parseCustomBoard :: String -> Int -> Board
parseCustomBoard input dim = Board {
    grid = listArray ((0, 0), (dim-1, dim-1)) discStates,
    boardDim = dim
  }
  where
    rows = lines input
    discStates = concatMap parseRow rows
    parseRow row = map parseDiscState (words row)
    parseDiscState "E" = Empty
    parseDiscState "R" = Placed Red
    parseDiscState "B" = Placed Blue
    parseDiscState _   = error "Invalid disc state in custom board. Use 'E', 'R', or 'B'."

-- Define a simple entry point for the program
-- Define a simple entry point for the program
main :: IO ()
main = do
    -- Get command-line arguments
    args <- getArgs

    -- Parse depth arguments
    let (depth, parallelDepth, inputFile) = case args of
            (d:p:file:_) -> (read d, read p, Just file)
            (d:p:_)      -> (read d, read p, Nothing)
            _            -> error "Usage: <depth> <parallelDepth> [gameboard file]"

    -- Load the custom board or use the default initial game state
    gameSetup <- case inputFile of
        Just file -> do
            content <- readFile file
            let customBoard = parseCustomBoard content 8 -- Assume an 8x8 board for simplicity
            return GameSetup {
                board = customBoard,
                aiPlays = [Red, Blue],
                searchDepth = depth
            }
        Nothing -> return (initialGameState 8 [Red, Blue] depth)

    putStrLn $ "Using minimax depth: " ++ show depth
    putStrLn $ "Parallelizing at depth: " ++ show parallelDepth

    let currentPlayer = Red

    -- Calculate the next move for the current player with the specified depth
    let moves = getAIMoveParallel gameSetup currentPlayer depth parallelDepth
    if null moves
        then putStrLn "No valid moves available."
        else do
            let selectedMove = head moves
            let updatedBoard = applyMove (board gameSetup) currentPlayer [selectedMove]
            putStrLn $ "Next move: " ++ show selectedMove
            putStrLn "Updated board state:"
            print updatedBoard

--minimax search tree
minimax :: Board -> Player -> Int -> Int -> Int -> Int -> Int
minimax board player depth alpha beta parallelDepth
    | depth == 0 || null possibleMoves = evaluateBoard board player -- Base case
    | depth > parallelDepth = sequentialMinimax alpha beta possibleMoves 
    | depth == parallelDepth = parallelMinimax alpha beta possibleMoves 
    | otherwise = sequentialMinimax alpha beta possibleMoves 
  where
    possibleMoves = movesForPlayer board player

    -- Sequential minimax
    sequentialMinimax :: Int -> Int -> [[Coordinate]] -> Int
    sequentialMinimax a b [] = if player == maximizingPlayer then a else b
    sequentialMinimax a b (move:moves)
        | player == maximizingPlayer = maximizing a b moves move
        | otherwise = minimizing a b moves move
      where
        maximizing a b moves move = sequentialMinimax (max a (minimax (applyMove board player move) (opposingPlayer player) (depth - 1) a b parallelDepth)) b moves
        minimizing a b moves move = sequentialMinimax a (min b (minimax (applyMove board player move) (opposingPlayer player) (depth - 1) a b parallelDepth)) moves

    -- Parallel minimax
    parallelMinimax :: Int -> Int -> [[Coordinate]] -> Int
    parallelMinimax a b moves =
        let 
            -- Number of threads available
            threads = numCapabilities

            bufferSize = max 1 (round (2 * fromIntegral threads))
            chunkSize = max 1 (length moves `div` (max 1 (round (fromIntegral threads * fromIntegral threads/3))))
            chunkedMoves = chunksOf chunkSize moves

            results = map (maximum . map evaluateMove) chunkedMoves `using` parBuffer bufferSize rdeepseq
        in if player == maximizingPlayer
           then maximum results
           else minimum results
      where
        -- Evaluates a single move
        evaluateMove move = minimax (applyMove board player move) (opposingPlayer player) (depth - 1) a b parallelDepth
    maximizingPlayer = Red -- Assume Red is the maximizing player

-- AI with parallelism and controlled depth
getAIMoveParallel :: GameSetup -> Player -> Int -> Int -> [Coordinate]
getAIMoveParallel setup player depth parallelDepth
    | null possibleMoves = [] -- No moves available
    | otherwise = fst $ maximumBy (comparing snd) evaluatedMoves
  where
    -- Get all possible moves for the current player
    possibleMoves = movesForPlayer (board setup) player

    -- Evaluate moves in parallel
    evaluatedMoves = map evaluateMove possibleMoves `using` parBuffer 2 rdeepseq
    evaluateMove move = (move, minimax (applyMove (board setup) player move) (opposingPlayer player) (depth - 1) minBound maxBound parallelDepth) `using` rdeepseq

evaluateBoard :: Board -> Player -> Int
evaluateBoard board player = scoreForPlayer - scoreForOpponent
  where
    scoreForPlayer = length [pos | (pos, state) <- assocs (grid board), state == Placed player]
    scoreForOpponent = length [pos | (pos, state) <- assocs (grid board), state == Placed (opposingPlayer player)]
