module Main where

import System.Environment (getArgs)
import System.IO (readFile)
import Data.Array (array, listArray, assocs)
import Data.Foldable (maximumBy)
import Data.Ord (comparing)
import Control.Parallel (par, pseq)
import Control.Parallel.Strategies (parList, parListChunk, rdeepseq, using,parMap)
import Othello.GameLogic (
    GameSetup(..), Player(..), initialGameState, Board(..), DiscState(..),
    movesForPlayer, applyMove, Coordinate, opposingPlayer
  )

-- Parse a custom board state from a string
parseCustomBoard :: String -> Int -> Board
parseCustomBoard input dim = Board {
    grid = listArray ((0, 0), (dim-1, dim-1)) discStates,
    boardDim = dim
  }
  where
    rows = lines input
    discStates = concatMap parseRow rows
    parseRow row = map parseDiscState (words row)
    parseDiscState "E" = Empty
    parseDiscState "R" = Placed Red
    parseDiscState "B" = Placed Blue
    parseDiscState _   = error "Invalid disc state in custom board. Use 'E', 'R', or 'B'."

-- Define a simple entry point for the program
main :: IO ()
main = do
    -- Get command-line arguments
    args <- getArgs

    -- Parse depth arguments
    let (depth, parallelDepth, inputFile) = case args of
            (d:p:file:_) -> (read d, read p, Just file)
            (d:p:_)      -> (read d, read p, Nothing)
            _            -> error "Usage: <depth> <parallelDepth> [gameboard file]"

    -- Load the custom board or use the default initial game state
    gameSetup <- case inputFile of
        Just file -> do
            content <- readFile file
            let customBoard = parseCustomBoard content 8 -- Assume an 8x8 board for simplicity
            return GameSetup {
                board = customBoard,
                aiPlays = [Red, Blue],
                searchDepth = depth
            }
        Nothing -> return (initialGameState 8 [Red, Blue] depth)

    putStrLn $ "Using minimax depth: " ++ show depth
    putStrLn $ "Parallelizing at depth: " ++ show parallelDepth

    let currentPlayer = Red

    -- Calculate the next move for the current player with the specified depth
    let moves = getAIMove gameSetup currentPlayer depth parallelDepth
    if null moves
        then putStrLn "No valid moves available."
        else do
            let selectedMove = head moves
            let updatedBoard = applyMove (board gameSetup) currentPlayer [selectedMove]
            putStrLn $ "Next move: " ++ show selectedMove
            putStrLn "Updated board state:"
            print updatedBoard

-- Minimax with optional parallelization
minimax :: Board -> Player -> Int -> Int -> Int -> Int -> Int
minimax board player depth alpha beta parallelDepth
    | depth == 0 || null possibleMoves = evaluateBoard board player -- Base case: evaluate board
    | depth == parallelDepth = parallelMinimax possibleMoves alpha beta
    | otherwise = sequentialMinimax possibleMoves alpha beta
  where
    possibleMoves = movesForPlayer board player
    maximizingPlayer = Red -- Assume Red is the maximizing player

    -- Parallel minimax evaluation (Option 1: Parallelize at one depth)
    parallelMinimax :: [[Coordinate]] -> Int -> Int -> Int
    parallelMinimax moves a b =
        let results = parMap rdeepseq (evaluateMove board player (depth - 1) a b) moves
        in maximum results

    -- Sequential minimax evaluation
    sequentialMinimax :: [[Coordinate]] -> Int -> Int -> Int
    sequentialMinimax [] a _ = a -- No more moves, return alpha
    sequentialMinimax (move:moves) a b
        | a >= b    = a -- Prune the rest of the tree
        | otherwise = sequentialMinimax moves a' b
      where
        a' = max a (evaluateMove board player (depth - 1) a b move)

    evaluateMove :: Board -> Player -> Int -> Int -> Int -> [Coordinate] -> Int
    evaluateMove b p d a b' move = minimax (applyMove b p move) (opposingPlayer p) d a b' parallelDepth

-- Parallelization Option 2: Divide the Tree Aggressively
parallelMinimaxAggressive :: Board -> Player -> Int -> Int -> Int -> [[Coordinate]] -> Int
parallelMinimaxAggressive board player depth a b moves =
    let chunkedMoves = map (:[]) moves `using` parListChunk 2 rdeepseq
        results = map (maximum . map (evaluateMove board player (depth - 1) a b)) chunkedMoves
    in maximum results
  where
    evaluateMove :: Board -> Player -> Int -> Int -> Int -> [Coordinate] -> Int
    evaluateMove b p d a b' move = minimax (applyMove b p move) (opposingPlayer p) d a b' depth

-- AI
getAIMove :: GameSetup -> Player -> Int -> Int -> [Coordinate]
getAIMove setup player depth parallelDepth
    | null possibleMoves = [] -- No moves available
    | otherwise = fst $ maximumBy (comparing snd) evaluatedMoves
  where
    -- Get all possible moves for the current player
    possibleMoves = movesForPlayer (board setup) player

    -- Evaluate each move using minimax
    evaluatedMoves = [
        (move, minimax (applyMove (board setup) player move) (opposingPlayer player) (depth - 1) minBound maxBound parallelDepth)
        | move <- possibleMoves
      ]

-- Example board evaluation function
evaluateBoard :: Board -> Player -> Int
evaluateBoard board player = scoreForPlayer - scoreForOpponent
  where
    scoreForPlayer = length [pos | (pos, state) <- assocs (grid board), state == Placed player]
    scoreForOpponent = length [pos | (pos, state) <- assocs (grid board), state == Placed (opposingPlayer player)]
