import subprocess
import sys
import re
import argparse

def parse_output(stdout):
    total_time_pattern = r"Total\s+time\s+\d+\.\d+s\s+\(\s+([\d\.]+)s" 
    tasks_pattern = r"TASKS:\s+(\d+)"  
    sparks_pattern = r"SPARKS:\s+(\d+)"

    total_time_match = re.search(total_time_pattern, stdout)
    tasks_match = re.search(tasks_pattern, stdout)
    sparks_match = re.search(sparks_pattern, stdout)

    total_time = float(total_time_match.group(1)) if total_time_match else None
    tasks = int(tasks_match.group(1)) if tasks_match else None
    sparks = int(sparks_match.group(1)) if sparks_match else None

    return total_time, tasks, sparks

def run_benchmark(test_file="custom_game_1.txt", iterate="parallel_depth", depth=7, parallel_depth=1):
    results = {}
    iteration_range = range(1, max(depth, parallel_depth) + 1)
    threads_range = range(1, 9) 

    if iterate == "depth":
        depth_range = iteration_range
        parallel_depth_range = [parallel_depth]
        threads_range = [4]  
    elif iterate == "parallel_depth":
        depth_range = [depth]
        parallel_depth_range = iteration_range
        threads_range = [4] 
    elif iterate == "num_threads":
        depth_range = [depth]
        parallel_depth_range = [parallel_depth]
    else:
        print(f"Invalid iterate value: {iterate}. Use 'depth', 'parallel_depth', or 'num_threads'.")
        return

    for d in depth_range:
        for pd in parallel_depth_range:
            for threads in threads_range:
                print(f"Running tests with depth {d}, parallel depth {pd}, threads {threads}...")
                
                total_times = []
                total_tasks = []
                total_sparks = []

                for _ in range(5):
                    # Prepare the command
                    command = [
                        "cabal", "run", "othello", "--", str(d), str(pd), test_file, "+RTS", "-s", f"-N{threads}"
                    ]
                    
                    try:
                        result = subprocess.run(command, capture_output=True, text=True, check=True)
                        stdout = result.stdout
                        stderr = result.stderr

                        total_time, tasks, sparks = parse_output(stdout + stderr)

                        if total_time is not None and tasks is not None and sparks is not None:
                            total_times.append(total_time)
                            total_tasks.append(tasks)
                            total_sparks.append(sparks)
                        else:
                            print(f"Warning: Missing data in the output for depth {d}, parallel depth {pd}, threads {threads}")
                        

                    except subprocess.CalledProcessError as e:
                        print(f"Error running command {command}: {e}")
                        continue
                
                # Compute averages for this configuration
                if total_times:
                    avg_total_time = sum(total_times) / len(total_times)
                    avg_tasks = sum(total_tasks) / len(total_tasks)
                    avg_sparks = sum(total_sparks) / len(total_sparks)
                    results[(d, pd, threads)] = {
                        "avg_total_time": avg_total_time,
                        "avg_tasks": avg_tasks,
                        "avg_sparks": avg_sparks
                    }
    
    for (d, pd, threads), data in results.items():
        print(f"Depth {d}, Parallel depth {pd}, Threads {threads}:")
        print(f"  Average Total Time: {data['avg_total_time']:.3f}s")
        print(f"  Average Tasks: {data['avg_tasks']}")
        print(f"  Average Sparks: {data['avg_sparks']}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run benchmarks for Othello.")
    parser.add_argument("--test_file", type=str, default="custom_game_1.txt", help="Path to the test file.")
    parser.add_argument("--iterate", type=str, choices=["depth", "parallel_depth", "num_threads"], default="parallel_depth",
                        help="Parameter to iterate over: 'depth', 'parallel_depth', or 'num_threads'.")
    parser.add_argument("--depth", type=int, default=7, help="Initial depth value.")
    parser.add_argument("--parallel_depth", type=int, default=1, help="Initial parallel depth value.")

    args = parser.parse_args()

    run_benchmark(
        test_file=args.test_file,
        iterate=args.iterate,
        depth=args.depth,
        parallel_depth=args.parallel_depth
    )
