{-#LANGUAGE InstanceSigs#-} -- Permit type declarations in instance definitions
{-# LANGUAGE TupleSections #-} -- Partial tuple constructors as functions
{-# LANGUAGE NamedFieldPuns #-} -- Allow more elegant construction of data

module Othello.GameLogic where
import Data.Array ( Array, array, elems, inRange, bounds, (//) )
import qualified Data.Array ((!))
import Data.Foldable (maximumBy, minimumBy)
import Data.Ord (comparing)
import Data.Array (Array, assocs, elems, bounds, (//))
-- Score is an integer, and coordinate is a pair of integers
type UnitScore = Int
type Coordinate = (Int, Int)

-- A player is either Red, or Blue. Derive comparison and Show
data Player = Red | Blue deriving (Eq, Show)

-- Each spot on a board is either empty, or placed with some player
data DiscState = Empty | Placed Player deriving (Eq, Show)

-- Board is essentially a grid of disc states, with the size attached
data Board = Board {
    grid :: Array Coordinate DiscState,
    boardDim :: Int
} deriving (Show)

-- For purposes of Minimax AI, we will need to measure the score of a given state. It will be either a win, indeterminate with score of some kind (from the view of who is requesting it), or a loss
data BoardScore = Win | Indeterminate UnitScore | Lose deriving (Eq, Show)

-- Define a order for a board score. For least complexity, define ordering as a set of comparative properties between different scores
instance Ord BoardScore where
    (<=) :: BoardScore -> BoardScore -> Bool
    (<=) Lose _ = True -- Lose is the smallest and definitely equal
    (<=) (Indeterminate _) Lose = False -- Indeterminate is never less or equal to a win
    (<=) (Indeterminate _) Win = True -- Indeterminate is always less than a win
    (<=) (Indeterminate a) (Indeterminate b) = a <= b -- For two indeterminates, their respective ordering depends on their scores
    (<=) Win Win = True -- Win is equal with a win
    (<=) Win _ = False -- Otherwise, no

-- Define a scoring function; given a player and a board, what is their score?
score :: Player -> Board -> BoardScore
score player board
    | not (movesPossibleOnBoard board) && redLeading = if player == Red then Win else Lose
    | not (movesPossibleOnBoard board) && blueLeading = if player == Blue then Win else Lose
    | otherwise = Indeterminate (if player == Red then redCount else blueCount)
    where
        redLeading = redCount > blueCount
        blueLeading = blueCount > redCount

        (redCount, blueCount) = pieceCount board

data GameSetup = GameSetup {
    board :: Board, -- Board
    aiPlays :: [Player], -- Which turns AI plays?
    searchDepth :: Int -- Search depth
} deriving (Show)
-- Core functions

-- Count of pieces, per color, for the board
pieceCount :: Board -> (Int, Int)
pieceCount board = foldr adder (0,0) (elems $ grid board) -- Add element by element, start with zero scores for both
    where
        adder :: DiscState -> (Int, Int) -> (Int, Int)
        adder state count@(red, blue) = case state of
            Empty -> count
            Placed Red -> (red+1, blue)
            Placed Blue -> (red, blue+1)

-- Definition of the opposing player for a given player
opposingPlayer :: Player -> Player
opposingPlayer Red = Blue
opposingPlayer Blue = Red

-- Calculating which player is winning by their score; this does not consider if there are more turns remaining
playerWithBestScore :: Board -> Maybe Player
playerWithBestScore board
    | red == blue = Nothing
    | otherwise = if red > blue then Just Red else Just Blue
    where
        (red, blue) = pieceCount board

-- Define an indexing operation for a board, quite alike what Arrays have
(!) :: Board -> Coordinate -> DiscState
board ! coordinate = (Data.Array.!) (grid board) coordinate

-- Define a validity check operator for indexes; this will return true if the index is acceptable
(!?) :: Board -> Coordinate -> Bool
board !? coord = inRange (bounds $ grid board) coord

-- Define a grid comprehension function; mapping over coordinates of a grid, construct some array of data
comprehensionByBoard :: Board -> (Coordinate -> x) -> [x]
comprehensionByBoard board = comprehensionByDim size
    where
        size = boardDim board

comprehensionByDim :: Int -> (Coordinate -> x) -> [x]
comprehensionByDim size func = [ func (a,b) | a <- [0..size-1], b <- [0..size-1]]

initialGameState :: Int -> [Player] -> Int -> GameSetup
initialGameState dim aiPlays searchDepth = GameSetup { aiPlays, searchDepth, board }
    where
        board = Board { boardDim = dim, grid = array ((0, 0), (dim-1,dim-1)) (comprehensionByDim dim (\p -> (p, startPieces p))) }

        startPieces :: Coordinate -> DiscState
        startPieces (cx, cy)
            | cx == (dim `div` 2) - 1 && cy == (dim `div` 2) - 1 = Placed Blue
            | cx == (dim `div` 2) && cy == (dim `div` 2) = Placed Blue
            | cx == (dim `div` 2) - 1 && cy == (dim `div` 2) = Placed Red
            | cx == (dim `div` 2) && cy == (dim `div` 2) - 1 = Placed Red
            | otherwise = Empty
---- Moves and AI

-- For a given board, player and coordinate, determine what coordinates should be changed to player's color. If an empty list is returned, move is not valid
-- This also includes the starting point given
getMovesOnPoint :: Board -> Player -> Coordinate -> [Coordinate]
getMovesOnPoint board player startCoord@(sx, sy)
    | not (board !? startCoord) = [] -- Coordinate is not a valid position
    | startPiece /= Empty = [] -- Starting piece is not empty
    | null resultSteps = [] -- No valid steps exist
    | otherwise = startCoord : resultSteps
    where
        resultSteps = concatMap walkAndMark directions -- Valid steps are a concatenation of walked directions - per rules, we can and must mark all branched paths

        walkAndMark :: Coordinate -> [Coordinate] -- If stepping from a given direction, what can we mark (excluding start position)?
        walkAndMark dir@(dx, dy) = walkAndMark' (sx+dx, sy+dy) dir []

        walkAndMark' :: Coordinate -> Coordinate -> [Coordinate] -> [Coordinate] -- Current position, direction, found already
        walkAndMark' cur@(cx, cy) dir@(dx, dy) found
            | isEndPiece = found -- End piece, terminate here
            | isValidTraversalPiece = walkAndMark' (cx+dx, cy+dy) dir (cur:found) -- Can traverse, step forward
            | otherwise = [] -- No valid way to travel nor end, return nothing
            where
                isEndPiece = isValidPos && board ! cur == Placed player
                isValidTraversalPiece = isValidPos && board ! cur == Placed (opposingPlayer player)
                isValidPos = board !? cur

        directions = [(-1,-1), (-1,0), (-1,1), (0,-1), (0,1), (1,-1), (1,0), (1,1)] -- Which directions need to be checked. As defined in the rules, we work on either horizontal, vertical or diagonal lines
        startPiece = board ! startCoord

movesForPlayer :: Board -> Player -> [[Coordinate]]  -- Return all movesets that are nonempty, for a given player?
movesForPlayer board player = filter (not . null) mappedCoords
    where
        mappedCoords = comprehensionByBoard board (getMovesOnPoint board player)

movesPossibleOnBoard :: Board -> Bool -- Are there any moves possible on board?
movesPossibleOnBoard board = not (null (movesForPlayer board Red) && null (movesForPlayer board Blue))

applyMove :: Board -> Player -> [Coordinate] -> Board
applyMove board player moveList = board { grid = grid board // map (, Placed player) moveList }

-- AI

getAIMove :: GameSetup -> Player -> Int -> [Coordinate]
getAIMove setup player depth
    | null possibleMoves = [] -- No moves available
    | otherwise = fst $ maximumBy (comparing snd) evaluatedMoves
  where
    -- Get all possible moves for the current player
    possibleMoves = movesForPlayer (board setup) player

    -- Evaluate each move using minimax
    evaluatedMoves = [(move, minimax (applyMove (board setup) player move) (opposingPlayer player) (depth - 1)) | move <- possibleMoves]

-- Minimax algorithm for game tree evaluation
minimax :: Board -> Player -> Int -> Int
minimax board player depth
    | depth == 0 || null possibleMoves = evaluateBoard board player -- Base case: evaluate board
    | player == maximizingPlayer = maximum [minimax (applyMove board player move) (opposingPlayer player) (depth - 1) | move <- possibleMoves]
    | otherwise = minimum [minimax (applyMove board player move) (opposingPlayer player) (depth - 1) | move <- possibleMoves]
  where
    possibleMoves = movesForPlayer board player
    maximizingPlayer = Red -- Assume Red is the maximizing player

-- Example board evaluation function
evaluateBoard :: Board -> Player -> Int
evaluateBoard board player = scoreForPlayer - scoreForOpponent
  where
    scoreForPlayer = length [pos | (pos, state) <- assocs (grid board), state == Placed player]
    scoreForOpponent = length [pos | (pos, state) <- assocs (grid board), state == Placed (opposingPlayer player)]
